@@ -1,13 +1,11 @@
 /* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
 
-#include "packets.h"
+#include "queueing.h"
 #include "socket.h"
 #include "timers.h"
 #include "device.h"
-#include "config.h"
 #include "ratelimiter.h"
 #include "peer.h"
-#include "uapi.h"
 #include "messages.h"
 
 #include <linux/module.h>
@@ -30,7 +28,7 @@ static int open(struct net_device *dev)
 	int ret;
 	struct wireguard_peer *peer, *temp;
 	struct wireguard_device *wg = netdev_priv(dev);
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
+#ifndef COMPAT_CANNOT_USE_IN6_DEV_GET
 	struct inet6_dev *dev_v6 = __in6_dev_get(dev);
 #endif
 	struct in_device *dev_v4 = __in_dev_get_rtnl(dev);
@@ -43,12 +41,12 @@ static int open(struct net_device *dev)
 		IN_DEV_CONF_SET(dev_v4, SEND_REDIRECTS, false);
 		IPV4_DEVCONF_ALL(dev_net(dev), SEND_REDIRECTS) = false;
 	}
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
+#ifndef COMPAT_CANNOT_USE_IN6_DEV_GET
 	if (dev_v6)
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)
-		dev_v6->addr_gen_mode = IN6_ADDR_GEN_MODE_NONE;
-#else
+#ifndef COMPAT_CANNOT_USE_DEV_CNF
 		dev_v6->cnf.addr_gen_mode = IN6_ADDR_GEN_MODE_NONE;
+#else
+		dev_v6->addr_gen_mode = IN6_ADDR_GEN_MODE_NONE;
 #endif
 #endif
 
@@ -56,8 +54,7 @@ static int open(struct net_device *dev)
 	if (ret < 0)
 		return ret;
 	peer_for_each (wg, peer, temp, true) {
-		timers_init_peer(peer);
-		packet_send_queue(peer);
+		packet_send_staged_packets(peer);
 		if (peer->persistent_keepalive_interval)
 			packet_send_keepalive(peer);
 	}
@@ -79,12 +76,11 @@ static int suspending_clear_noise_peers(struct notifier_block *nb, unsigned long
 			noise_handshake_clear(&peer->handshake);
 			noise_keypairs_clear(&peer->keypairs);
 			if (peer->timers_enabled)
-				del_timer(&peer->timer_kill_ephemerals);
+				del_timer(&peer->timer_zero_key_material);
 		}
 	}
 	rtnl_unlock();
 	rcu_barrier_bh();
-
 	return 0;
 }
 static struct notifier_block clear_peers_on_suspend = { .notifier_call = suspending_clear_noise_peers };
@@ -94,12 +90,12 @@ static int stop(struct net_device *dev)
 {
 	struct wireguard_device *wg = netdev_priv(dev);
 	struct wireguard_peer *peer, *temp;
+
 	peer_for_each (wg, peer, temp, true) {
-		timers_uninit_peer(peer);
+		skb_queue_purge(&peer->staged_packet_queue);
+		timers_stop(peer);
 		noise_handshake_clear(&peer->handshake);
 		noise_keypairs_clear(&peer->keypairs);
-		if (peer->timers_enabled)
-			del_timer(&peer->timer_kill_ephemerals);
 	}
 	skb_queue_purge(&wg->incoming_handshakes);
 	socket_uninit(wg);
@@ -111,6 +107,7 @@ static netdev_tx_t xmit(struct sk_buff *skb, struct net_device *dev)
 	struct wireguard_device *wg = netdev_priv(dev);
 	struct wireguard_peer *peer;
 	struct sk_buff *next;
+	struct sk_buff_head packets;
 	int ret;
 
 	if (unlikely(dev_recursion_level() > 4)) {
@@ -141,11 +138,7 @@ static netdev_tx_t xmit(struct sk_buff *skb, struct net_device *dev)
 		goto err_peer;
 	}
 
-	/* If the queue is getting too big, we start removing the oldest packets until it's small again.
-	 * We do this before adding the new packet, so we don't remove GSO segments that are in excess. */
-	while (skb_queue_len(&peer->tx_packet_queue) > MAX_QUEUED_OUTGOING_PACKETS)
-		dev_kfree_skb(skb_dequeue(&peer->tx_packet_queue));
-
+	__skb_queue_head_init(&packets);
 	if (!skb_is_gso(skb))
 		skb->next = NULL;
 	else {
@@ -169,10 +162,19 @@ static netdev_tx_t xmit(struct sk_buff *skb, struct net_device *dev)
 		 * so at this point we're in a position to drop it. */
 		skb_dst_drop(skb);
 
-		skb_queue_tail(&peer->tx_packet_queue, skb);
+		__skb_queue_tail(&packets, skb);
 	} while ((skb = next) != NULL);
 
-	packet_send_queue(peer);
+	spin_lock_bh(&peer->staged_packet_queue.lock);
+	/* If the queue is getting too big, we start removing the oldest packets until it's small again.
+	 * We do this before adding the new packet, so we don't remove GSO segments that are in excess. */
+	while (skb_queue_len(&peer->staged_packet_queue) > MAX_STAGED_PACKETS)
+		dev_kfree_skb(__skb_dequeue(&peer->staged_packet_queue));
+	skb_queue_splice_tail(&packets, &peer->staged_packet_queue);
+	spin_unlock_bh(&peer->staged_packet_queue.lock);
+
+	packet_send_staged_packets(peer);
+
 	peer_put(peer);
 	return NETDEV_TX_OK;
 
@@ -188,28 +190,11 @@ err:
 	return ret;
 }
 
-static int ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
-{
-	struct wireguard_device *wg = netdev_priv(dev);
-
-	if (!ns_capable(dev_net(dev)->user_ns, CAP_NET_ADMIN))
-		return -EPERM;
-
-	switch (cmd) {
-	case WG_GET_DEVICE:
-		return config_get_device(wg, ifr->ifr_ifru.ifru_data);
-	case WG_SET_DEVICE:
-		return config_set_device(wg, ifr->ifr_ifru.ifru_data);
-	}
-	return -EINVAL;
-}
-
 static const struct net_device_ops netdev_ops = {
 	.ndo_open		= open,
 	.ndo_stop		= stop,
 	.ndo_start_xmit		= xmit,
-	.ndo_get_stats64	= ip_tunnel_get_stats64,
-	.ndo_do_ioctl		= ioctl
+	.ndo_get_stats64	= ip_tunnel_get_stats64
 };
 
 static void destruct(struct net_device *dev)
@@ -220,16 +205,14 @@ static void destruct(struct net_device *dev)
 	list_del(&wg->device_list);
 	rtnl_unlock();
 	mutex_lock(&wg->device_update_lock);
-	peer_remove_all(wg);
+	peer_remove_all(wg); /* The final references are cleared in the below calls to destroy_workqueue. */
 	wg->incoming_port = 0;
-	destroy_workqueue(wg->incoming_handshake_wq);
-	destroy_workqueue(wg->peer_wq);
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	padata_free(wg->encrypt_pd);
-	padata_free(wg->decrypt_pd);
-	destroy_workqueue(wg->crypt_wq);
-#endif
-	routing_table_free(&wg->peer_routing_table);
+	destroy_workqueue(wg->handshake_receive_wq);
+	destroy_workqueue(wg->handshake_send_wq);
+	packet_queue_free(&wg->decrypt_queue, true);
+	packet_queue_free(&wg->encrypt_queue, true);
+	destroy_workqueue(wg->packet_crypt_wq);
+	routing_table_free(&wg->peer_routing_table, &wg->device_update_lock);
 	ratelimiter_uninit();
 	memzero_explicit(&wg->static_identity, sizeof(struct noise_static_identity));
 	skb_queue_purge(&wg->incoming_handshakes);
@@ -240,9 +223,7 @@ static void destruct(struct net_device *dev)
 	put_net(wg->creating_net);
 
 	pr_debug("%s: Interface deleted\n", dev->name);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
 	free_netdev(dev);
-#endif
 }
 
 static void setup(struct net_device *dev)
@@ -251,20 +232,14 @@ static void setup(struct net_device *dev)
 	enum { WG_NETDEV_FEATURES = NETIF_F_HW_CSUM | NETIF_F_RXCSUM | NETIF_F_SG | NETIF_F_GSO | NETIF_F_GSO_SOFTWARE | NETIF_F_HIGHDMA };
 
 	dev->netdev_ops = &netdev_ops;
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
-	dev->destructor = destruct;
-#else
-	dev->priv_destructor = destruct;
-	dev->needs_free_netdev = true;
-#endif
 	dev->hard_header_len = 0;
 	dev->addr_len = 0;
 	dev->needed_headroom = DATA_PACKET_HEAD_ROOM;
 	dev->needed_tailroom = noise_encrypted_len(MESSAGE_PADDING_MULTIPLE);
 	dev->type = ARPHRD_NONE;
 	dev->flags = IFF_POINTOPOINT | IFF_NOARP;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 3, 0)
-	dev->flags |= IFF_NO_QUEUE;
+#ifndef COMPAT_CANNOT_USE_IFF_NO_QUEUE
+	dev->priv_flags |= IFF_NO_QUEUE;
 #else
 	dev->tx_queue_len = 0;
 #endif
@@ -278,11 +253,12 @@ static void setup(struct net_device *dev)
 	netif_keep_dst(dev);
 
 	memset(wg, 0, sizeof(struct wireguard_device));
+	wg->dev = dev;
 }
 
-static int newlink(struct net *src_net, struct net_device *dev, struct nlattr *tb[], struct nlattr *data[])
+static int newlink(struct net *src_net, struct net_device *dev, struct nlattr *tb[], struct nlattr *data[], struct netlink_ext_ack *extack)
 {
-	int ret = -ENOMEM, cpu;
+	int ret = -ENOMEM;
 	struct wireguard_device *wg = netdev_priv(dev);
 
 	wg->creating_net = get_net(src_net);
@@ -295,73 +271,63 @@ static int newlink(struct net *src_net, struct net_device *dev, struct nlattr *t
 	routing_table_init(&wg->peer_routing_table);
 	cookie_checker_init(&wg->cookie_checker, wg);
 	INIT_LIST_HEAD(&wg->peer_list);
+	wg->device_update_gen = 1;
 
 	dev->tstats = netdev_alloc_pcpu_stats(struct pcpu_sw_netstats);
 	if (!dev->tstats)
 		goto error_1;
 
-	wg->incoming_handshakes_worker = alloc_percpu(struct handshake_worker);
+	wg->incoming_handshakes_worker = packet_alloc_percpu_multicore_worker(packet_handshake_receive_worker, wg);
 	if (!wg->incoming_handshakes_worker)
 		goto error_2;
-	for_each_possible_cpu (cpu) {
-		per_cpu_ptr(wg->incoming_handshakes_worker, cpu)->wg = wg;
-		INIT_WORK(&per_cpu_ptr(wg->incoming_handshakes_worker, cpu)->work, packet_process_queued_handshake_packets);
-	}
-	atomic_set(&wg->incoming_handshake_seqnr, 0);
 
-	wg->incoming_handshake_wq = alloc_workqueue("wg-kex-%s", WQ_CPU_INTENSIVE | WQ_FREEZABLE, 0, dev->name);
-	if (!wg->incoming_handshake_wq)
+	wg->handshake_receive_wq = alloc_workqueue("wg-kex-%s", WQ_CPU_INTENSIVE | WQ_FREEZABLE, 0, dev->name);
+	if (!wg->handshake_receive_wq)
 		goto error_3;
 
-	wg->peer_wq = alloc_workqueue("wg-kex-%s", WQ_UNBOUND | WQ_FREEZABLE, 0, dev->name);
-	if (!wg->peer_wq)
+	wg->handshake_send_wq = alloc_workqueue("wg-kex-%s", WQ_UNBOUND | WQ_FREEZABLE, 0, dev->name);
+	if (!wg->handshake_send_wq)
 		goto error_4;
 
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	wg->crypt_wq = alloc_workqueue("wg-crypt-%s", WQ_CPU_INTENSIVE | WQ_MEM_RECLAIM, 2, dev->name);
-	if (!wg->crypt_wq)
+	wg->packet_crypt_wq = alloc_workqueue("wg-crypt-%s", WQ_CPU_INTENSIVE | WQ_MEM_RECLAIM, 0, dev->name);
+	if (!wg->packet_crypt_wq)
 		goto error_5;
 
-	wg->encrypt_pd = padata_alloc_possible(wg->crypt_wq);
-	if (!wg->encrypt_pd)
+	if (packet_queue_init(&wg->encrypt_queue, packet_encrypt_worker, true, MAX_QUEUED_PACKETS) < 0)
 		goto error_6;
-	padata_start(wg->encrypt_pd);
 
-	wg->decrypt_pd = padata_alloc_possible(wg->crypt_wq);
-	if (!wg->decrypt_pd)
+	if (packet_queue_init(&wg->decrypt_queue, packet_decrypt_worker, true, MAX_QUEUED_PACKETS) < 0)
 		goto error_7;
-	padata_start(wg->decrypt_pd);
-#endif
 
 	ret = ratelimiter_init();
 	if (ret < 0)
 		goto error_8;
 
 	ret = register_netdevice(dev);
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
 	if (ret < 0)
 		goto error_9;
-#endif
+
 	list_add(&wg->device_list, &device_list);
+
+	/* We wait until the end to assign priv_destructor, so that register_netdevice doesn't
+	 * call it for us if it fails. */
+	dev->priv_destructor = destruct;
+
 	pr_debug("%s: Interface created\n", dev->name);
 	return ret;
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
 error_9:
 	ratelimiter_uninit();
-#endif
 error_8:
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	padata_free(wg->decrypt_pd);
+	packet_queue_free(&wg->decrypt_queue, true);
 error_7:
-	padata_free(wg->encrypt_pd);
+	packet_queue_free(&wg->encrypt_queue, true);
 error_6:
-	destroy_workqueue(wg->crypt_wq);
+	destroy_workqueue(wg->packet_crypt_wq);
 error_5:
-#endif
-	destroy_workqueue(wg->peer_wq);
+	destroy_workqueue(wg->handshake_send_wq);
 error_4:
-	destroy_workqueue(wg->incoming_handshake_wq);
+	destroy_workqueue(wg->handshake_receive_wq);
 error_3:
 	free_percpu(wg->incoming_handshakes_worker);
 error_2:
@@ -382,13 +348,14 @@ int __init device_init(void)
 {
 #ifdef CONFIG_PM_SLEEP
 	int ret = register_pm_notifier(&clear_peers_on_suspend);
+
 	if (ret)
 		return ret;
 #endif
 	return rtnl_link_register(&link_ops);
 }
 
-void __exit device_uninit(void)
+void device_uninit(void)
 {
 	rtnl_link_unregister(&link_ops);
 #ifdef CONFIG_PM_SLEEP
