@@ -3,7 +3,7 @@
 #include "device.h"
 #include "peer.h"
 #include "socket.h"
-#include "packets.h"
+#include "queueing.h"
 #include "messages.h"
 
 #include <linux/ctype.h>
@@ -30,6 +30,7 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 
 	skb->next = skb->prev = NULL;
 	skb->dev = wg->dev;
+	skb->mark = wg->fwmark;
 
 	rcu_read_lock_bh();
 	sock = rcu_dereference_bh(wg->sock4);
@@ -44,9 +45,14 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 
 	if (!rt) {
 		security_sk_classify_flow(sock, flowi4_to_flowi(&fl));
+		if (unlikely(!inet_confirm_addr(sock_net(sock), NULL, 0, fl.saddr, RT_SCOPE_HOST))) {
+			endpoint->src4.s_addr = *(__force __be32 *)&endpoint->src_if4 = fl.saddr = 0;
+			if (cache)
+				dst_cache_reset(cache);
+		}
 		rt = ip_route_output_flow(sock_net(sock), &fl, sock);
-		if (unlikely(endpoint->src4.s_addr && ((IS_ERR(rt) && PTR_ERR(rt) == -EINVAL) || (!IS_ERR(rt) && !inet_confirm_addr(sock_net(sock), rcu_dereference_bh(rt->dst.dev->ip_ptr), 0, fl.saddr, RT_SCOPE_HOST))))) {
-			endpoint->src4.s_addr = fl.saddr = 0;
+		if (unlikely(endpoint->src_if4 && ((IS_ERR(rt) && PTR_ERR(rt) == -EINVAL) || (!IS_ERR(rt) && rt->dst.dev->ifindex != endpoint->src_if4)))) {
+			endpoint->src4.s_addr = *(__force __be32 *)&endpoint->src_if4 = fl.saddr = 0;
 			if (cache)
 				dst_cache_reset(cache);
 			if (!IS_ERR(rt))
@@ -95,6 +101,7 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 
 	skb->next = skb->prev = NULL;
 	skb->dev = wg->dev;
+	skb->mark = wg->fwmark;
 
 	rcu_read_lock_bh();
 	sock = rcu_dereference_bh(wg->sock6);
@@ -161,8 +168,10 @@ int socket_send_skb_to_peer(struct wireguard_peer *peer, struct sk_buff *skb, u8
 int socket_send_buffer_to_peer(struct wireguard_peer *peer, void *buffer, size_t len, u8 ds)
 {
 	struct sk_buff *skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
+
 	if (unlikely(!skb))
 		return -ENOMEM;
+
 	skb_reserve(skb, SKB_HEADER_LEN);
 	memcpy(skb_put(skb, len), buffer, len);
 	return socket_send_skb_to_peer(peer, skb, ds);
@@ -196,7 +205,7 @@ int socket_send_buffer_as_reply_to_skb(struct wireguard_device *wg, struct sk_bu
 	return ret;
 }
 
-int socket_endpoint_from_skb(struct endpoint *endpoint, struct sk_buff *skb)
+int socket_endpoint_from_skb(struct endpoint *endpoint, const struct sk_buff *skb)
 {
 	memset(endpoint, 0, sizeof(struct endpoint));
 	if (skb->protocol == htons(ETH_P_IP)) {
@@ -204,51 +213,57 @@ int socket_endpoint_from_skb(struct endpoint *endpoint, struct sk_buff *skb)
 		endpoint->addr4.sin_port = udp_hdr(skb)->source;
 		endpoint->addr4.sin_addr.s_addr = ip_hdr(skb)->saddr;
 		endpoint->src4.s_addr = ip_hdr(skb)->daddr;
+		endpoint->src_if4 = skb->skb_iif;
 	} else if (skb->protocol == htons(ETH_P_IPV6)) {
 		endpoint->addr6.sin6_family = AF_INET6;
 		endpoint->addr6.sin6_port = udp_hdr(skb)->source;
 		endpoint->addr6.sin6_addr = ipv6_hdr(skb)->saddr;
 		endpoint->addr6.sin6_scope_id = ipv6_iface_scope_id(&ipv6_hdr(skb)->saddr, skb->skb_iif);
-		/* TODO: endpoint->addr6.sin6_flowinfo */
 		endpoint->src6 = ipv6_hdr(skb)->daddr;
 	} else
 		return -EINVAL;
 	return 0;
 }
 
-void socket_set_peer_endpoint(struct wireguard_peer *peer, struct endpoint *endpoint)
+static inline bool endpoint_eq(const struct endpoint *a, const struct endpoint *b)
 {
+	return (a->addr.sa_family == AF_INET && b->addr.sa_family == AF_INET &&
+		a->addr4.sin_port == b->addr4.sin_port && a->addr4.sin_addr.s_addr == b->addr4.sin_addr.s_addr &&
+		a->src4.s_addr == b->src4.s_addr && a->src_if4 == b->src_if4) ||
+	       (a->addr.sa_family == AF_INET6 && b->addr.sa_family == AF_INET6 &&
+		a->addr6.sin6_port == b->addr6.sin6_port && ipv6_addr_equal(&a->addr6.sin6_addr, &b->addr6.sin6_addr) &&
+		a->addr6.sin6_scope_id == b->addr6.sin6_scope_id && ipv6_addr_equal(&a->src6, &b->src6)) ||
+	       unlikely(!a->addr.sa_family && !b->addr.sa_family);
+}
+
+void socket_set_peer_endpoint(struct wireguard_peer *peer, const struct endpoint *endpoint)
+{
+	/* First we check unlocked, in order to optimize, since it's pretty rare
+	 * that an endpoint will change. If we happen to be mid-write, and two
+	 * CPUs wind up writing the same thing or something slightly different,
+	 * it doesn't really matter much either. */
+	if (endpoint_eq(endpoint, &peer->endpoint))
+		return;
+	write_lock_bh(&peer->endpoint_lock);
 	if (endpoint->addr.sa_family == AF_INET) {
-		read_lock_bh(&peer->endpoint_lock);
-		if (likely(peer->endpoint.addr4.sin_family == AF_INET &&
-			   peer->endpoint.addr4.sin_port == endpoint->addr4.sin_port &&
-			   peer->endpoint.addr4.sin_addr.s_addr == endpoint->addr4.sin_addr.s_addr &&
-			   peer->endpoint.src4.s_addr == endpoint->src4.s_addr))
-			goto out;
-		read_unlock_bh(&peer->endpoint_lock);
-		write_lock_bh(&peer->endpoint_lock);
 		peer->endpoint.addr4 = endpoint->addr4;
 		peer->endpoint.src4 = endpoint->src4;
+		peer->endpoint.src_if4 = endpoint->src_if4;
 	} else if (endpoint->addr.sa_family == AF_INET6) {
-		read_lock_bh(&peer->endpoint_lock);
-		if (likely(peer->endpoint.addr6.sin6_family == AF_INET6 &&
-			   peer->endpoint.addr6.sin6_port == endpoint->addr6.sin6_port &&
-			   /* TODO: peer->endpoint.addr6.sin6_flowinfo == endpoint->addr6.sin6_flowinfo && */
-			   ipv6_addr_equal(&peer->endpoint.addr6.sin6_addr, &endpoint->addr6.sin6_addr) &&
-			   peer->endpoint.addr6.sin6_scope_id == endpoint->addr6.sin6_scope_id &&
-			   ipv6_addr_equal(&peer->endpoint.src6, &endpoint->src6)))
-			goto out;
-		read_unlock_bh(&peer->endpoint_lock);
-		write_lock_bh(&peer->endpoint_lock);
 		peer->endpoint.addr6 = endpoint->addr6;
 		peer->endpoint.src6 = endpoint->src6;
 	} else
-		return;
+		goto out;
 	dst_cache_reset(&peer->endpoint_cache);
-	write_unlock_bh(&peer->endpoint_lock);
-	return;
 out:
-	read_unlock_bh(&peer->endpoint_lock);
+	write_unlock_bh(&peer->endpoint_lock);
+}
+
+void socket_set_peer_endpoint_from_skb(struct wireguard_peer *peer, const struct sk_buff *skb)
+{
+	struct endpoint endpoint;
+	if (!socket_endpoint_from_skb(&endpoint, skb))
+		socket_set_peer_endpoint(peer, &endpoint);
 }
 
 void socket_clear_peer_endpoint_src(struct wireguard_peer *peer)
@@ -318,6 +333,7 @@ int socket_init(struct wireguard_device *wg)
 		.ipv6_v6only = true
 	};
 #endif
+
 	mutex_lock(&wg->socket_update_lock);
 #if IS_ENABLED(CONFIG_IPV6)
 retry:
@@ -365,6 +381,7 @@ out:
 void socket_uninit(struct wireguard_device *wg)
 {
 	struct sock *old4, *old6;
+
 	mutex_lock(&wg->socket_update_lock);
 	old4 = rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock));
 	old6 = rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock));
